/*==========================================================================
  Filename: Cango-5v19.js
  Rev: 5
  By: A.R.Collins
  Description: A graphics library for the canvas element.
  License: Released into the public domain
  latest version at
  <http://www/arc.id.au/CanvasGraphics.html>
  Report bugs to tony at arc.id.au

  Date   |Description                                          |By
  --------------------------------------------------------------------------
  14Oct12 Rev 1.00 First release based on Cango0v43             ARC
  23Oct12 bugfix: xRef,yRef not scaled in most compile methods
          don't pre-scale xRef, yRef in compile methods         ARC
  24Oct12 Give Obj2D methods to simplify compile methods        ARC
  25Oct12 Eliminated group etc. Just use arrays
          Renamed movePoints etc to translate, rotate, scale
          Pass render arguments to color functions              ARC
  26Oct12 Restored the xRef,yRef parameters to compileSvg cmds
          bugfix: restore scaling of font weight                ARC
  27Oct12 Removed iso commands - use 'iso' parameter instead
          bugfix: obj.translate and compileSvg used opposite
          sign conventions - switch compileSvg                  ARC
  11Nov12 removed Object.clone to avoid conflicts with 3rd
          party code (eg ACE editor)                            ARC
  19Nov12 Added to toWorldCoords                                ARC
  20Nov12 Added Obj2D.path to hold 'as rendered' coords
          Added Obj2D.dragNdrop, to hold Drag2D obj             ARC
  24Nov12 Added support for dragNdrop for rotated images
          Removed setRotation clearRotation capability          ARC
  25Nov12 Re-wrote paintPath to better use position transform
          Drag event handlers called in scope of dragNdrop obj
          Re-wrote paintImg to clarify draw origin offsets      ARC
  26Nov12 Added dragNdrop capability to TEXT objects            ARC
  27Nov12 Added Drag2D object, changed .path to .pxOutline
          Added 'as rendered' pxOrg (needed by enableDrag)      ARC
  29Nov11 Added parent as a Drag2D property                     ARC
          Released as Cango2v00                                 ARC
  30Nov12 Added Obj2D.setStrokeWidth
          Made drawPath scl factor permanent, used Obj2D.scale
          Made drawShape scl factor permanent, used Obj2D.scale ARC
  20Dec12 Enabled compileShape to accept numeric data           ARC
  02Jan13 Added stopAnimation,
          Removed animTranslate, animRotate, animScale          ARC
  13Jan13 Changed the shapeDefs to match Cango3D                ARC
  24Jan13 Added cross and ex to shapeDefs                       ARC
  31Jan13 bugfix: multiple Cango contexts over-writing
          mousedown event handler, fixed with global _draggable ARC
  16Feb13 Added Obj2D.appendPath and Obj2D.revWinding           ARC
  19Feb13 Added Obj2D.clipPath                                  ARC
  26Feb13 Use off-screen buffer and bitblt to screen canvas     ARC
  27Feb13 Added shim for isArray, and used it throughout        ARC
  28Feb13 bugfix: bitblt buffer after clearCanvas etc           ARC
  01Mar13 Allow optional direct drawing if node passed in
          Remove useless clearRect before bitblt                ARC
  02Mar13 Added support for multiple animations in one context  ARC
  03Mar13 Added deleteAnimation and uniqueId methods
          reverted to non-buffered drawing                      ARC
  04Mar13 Added pause and step to animations                    ARC
  05Mar13 Added setStepTIme                                     ARC
  06Mar13 Buffer animation drawing and bitblt completed frames  ARC
  10Mar13 Use setProperty and setPropertyDefault methods        ARC
  11Mar13 Added scl parameter to compilePath and compileShape
          Added lineWidth as parameter to compilePath
          dropped strokeColor parm from drawShape
          changed order of parms in drawPath and drawShape
          Added fontWeight and color as compileText parms       ARC
  12Mar13 Added fontSize and fontWeight to settable properties  ARC
  27Mar13 Use native canvasText                                 ARC
  02Apr13 Give clearCanvas an optional fill color               ARC
  28Apr13 bugfix: compile 's' command missing +x, +y            Ales Seifert
  01May13 Make strokeWidth scale with any path scaling          ARC
  31May13 bugfix: typo in clipPath                              ARC
  01Jun13 Use "TEXT","IMG","PATH","SHAPE" instead of objType
          bugfix: _resized etc should be Objects not Arrays
          Many style changes to make JSLint happy               ARC
  02Jun13 Cursor position and csr offsets all in world coords
          bugfix: no check if draggable already in _draggables  ARC
  03Jun13 Tidy render code                                      ARC
  06Jun13 bugfix: enableDrag had bad toWorldCoords call and
          reference to invalid pxOrgX and pxOrgY.
          Cleanup Cango mousedown handler                       ARC
  23Jul13 Re-write based on Cango2v38 and features from Cango2D ARC
  26Jul13 Removed drag as a parameter from methods              ARC
  27Jul13 Add createLayer and deleteLayer methods               ARC
  01Aug13 Put Animation stuff back in (was to be out of Ver3)   ARC
  03Aug13 Added support for Sprites                             ARC
  04Aug13 Add createSprite and _paintSprite                     ARC
  05Aug13 Make Sprite an object with gc and buf properties
          bugfix: bad test for animation.id in deleteAnimation  ARC
  06Aug13 bugfix: img, text, sprite bounding box not scaled
          by render scale factor
          Remove sprites, use off screen drawing and bitblt     ARC
  07Jun13 Add toImgObj
          bugfix: img rotation convention in wrong direction    ARC
  08Aug13 Change setWorldCoords parameters to
          llx, lly, Xspan, Yspan to match Cango2D and Cango3D
          bugfix: txt rotation convention in wrong direction    ARC
  09Aug13 Switch Xspan Yspan to spanX spanY to match Cango3D
          bugfix: rAF not passing good time, use Date.now       ARC
  12Aug13 Make rAF shim separate namespace (function)           ARC
  19Mar14 render to accept Obj2D or array of Obj2D and now
          array elements may be nested arrays of Obj2D etc      ARC
  20Mar14 bugfix: text imgX,imgY not scaling with font          ARC
  02Apr14 bugfix: transforms using invalid this.xscl,this.yscl
          (bug??) Make all Obj2D transforms 'iso'               ARC
  13Apr14 Allow -ve spanY values to give  Y +ve down the screen ARC
  27Apr14 Added validity check on lorg before its used as index
          Clean up to make JsLint happier                       ARC
  05May14 bugfix: Obj2D bad vpLL and offset parms in bbBox calc
          in translate, rotate and scale methods                ARC
  06May14 Ver 4: remove Obj2D factory fn references to Cango    ARC
  10May14 Allow all Obj2D properties to be set on creation
          Make default Obj2D type = PATH                        ARC
  11May14 Moved animation stuff to a seperate module
          Make shapeDefs function with parameters
          Remove scl parmater from compilePath & compileShape   ARC
  12May14 Make cgo2DtoDrawCmd an Obj2D method                   ARC
  13May14 Do fallback to default lineWidth, lineCap in render   ARC
          bugfix:render not handling a color function correctly ARC
  15May14 bugfix: image and text configure should be done at
          render time not compile to catch property changes     ARC
  16May14 Added imgWidth, imgHeight, lorg to setProperty
          Added imgLorgX, imgLorgY properties to Obj2D          ARC
  26May14 Renamed global Obj2D to Cobj and Drag2D to DnD to
          avoid clash with Cango2D globals                      ARC
  29May14 Added ellipse to shapeDefs
          Use setProperty in constructor not duplicated code    ARC
  30May14 Drop fontWeight as compileText argument
          Change font to Inconsolata (Inconsolata-dz)           ARC
  30Jun14 Tidy return value to keep Mozilla happy               ARC
  06Jul14 Added cancelDrag as a method of DnD object            ARC
  13Jul14 compileShape only take 1 color parameter: fillColor.
          bugfix: bad test for Cobj.type in clipPath            ARC
  14Jul14 Add Cobj.dropShadow method
          Released as Cango-5v00                                ARC
  16Jul14 Add Cobj.border (boolean) property to Cobj
          TEXT fill with fillColor and border with strokeColor
          bugfix: missing break in setProperty switch statement
          Add Cobj.setBorder method
          Renamed dropShadow method to setDropShadow            ARC
  17Jul14 bugfix: reset dropShadow before stroking text border  ARC
  18Jul14 Upgrade Cobj.appendPath to clone the drawCmd array    ARC
  19Jul14 Handle onmouseout to be same as onmouseup in DnD      ARC
  20Jul14 Added resize handler for overlay canvases.
          Make buf, draggables properties of each canvas        ARC
  24Jul14 Inhibit overlays creating layers
          Simplify render, drop functions returning color
          Synchronise animation timeline across all layers      ARC
  25Jul14 Added deleteAllLayers method                          ARC
  26Jul14 Use Layer object per layer to store draggables
          Update DnD layer to match the canvas used by render
          Make DnD private, pass handlers to enableDrag         ARC
  27Jul16 Split DnD off into separate module                    ARC
  31Jul14 bugfix: layers not tracking bkgCanvas top,left        ARC
  04Aug14 bugfix: TEXT type Cobj.translate not moving 'iso'
          bugfix: IMG type Cobj.translate Y coord inverted      ARC
  ==========================================================================*/

var Cobj, svgToCgo2D, shapeDefs;

var Cango = (function()
{
  "use strict";

  var uniqueVal = 0;    // used to generate unique value for different Cango instances

  var isArray = function(obj)
  {
    return Object.prototype.toString.call(obj) === '[object Array]';
  };

  var isNumber = function(o)
  {
    return !isNaN(o) && o !== null && o !== "" && o !== false;
  };

  // simple add event handler that has the handlers called in the sequence that they were set
  var addLoadEvent = function(obj, func)
  {
  	var oldonload = obj.onload;

  	if (typeof(obj.onload) != "function")
    {
      obj.onload = func;
    }
  	else
    {
    	obj.onload = function(){ oldonload(); func(); };
    }
  };

  var addEvent = function(element, eventType, handler)
  {
    if (element.attachEvent)
    {
     return element.attachEvent('on'+eventType, handler);
    }
    return element.addEventListener(eventType, handler, true);
  };

  var removeEvent = function(element, eventType, handler)
  {
   if (element.removeEventListener)
   {
      element.removeEventListener (eventType, handler, false);
   }
   if (element.detachEvent)
   {
      element.detachEvent ('on'+eventType, handler);
   }
  };

  if (!Array.prototype.contains)
  {
    Array.prototype.contains = function(obj)
    {
      var i = this.length;
      while (i--)
      {
        if (this[i] === obj)
        {
          return true;
        }
      }
      return false;
    };
  }

  if (!Array.prototype.indexOf)
  {
    Array.prototype.indexOf= function(find, i /*opt*/)
    {
      var n;
      if (i===undefined)
      {
        i= 0;
      }
      if (i<0)
      {
        i+= this.length;
      }
      if (i<0)
      {
        i= 0;
      }
      for (n = this.length; i < n; i++)
      {
        if (i in this && this[i]===find)
        {
          return i;
        }
      }
      return -1;
    };
  }

  if (!Array.prototype.map)
  {
    Array.prototype.map = function(fun, thisArg)
    {
      var i,
          len = this.length,
          res = [],
          thisp = thisArg;

      if (typeof fun != "function")
      {
        throw new TypeError();
      }
      for (i = 0; i < len; i++)
      {
        if (this.hasOwnProperty(i))
        {
          res[i] = fun.call(thisp, this[i], i, this);
        }
      }
      return res;
    };
  }

  function clone(orgItem)
  {
    var newItem = (isArray(orgItem)) ? [] : {},
        i;
    for (i in orgItem)
    {
      if (orgItem[i] && typeof orgItem[i] == "object")
      {
        newItem[i] = clone(orgItem[i]);
      }
      else
      {
        newItem[i] = orgItem[i];
      }
    }
    return newItem;
  }

  if (typeof shapeDefs == "undefined")
  {
    shapeDefs = {'circle': function(diameter){
                            var d = diameter || 1;
                            return ["M", -0.5*d,0,
                            "C", -0.5*d, -0.27614*d, -0.27614*d, -0.5*d, 0, -0.5*d,
                            "C", 0.27614*d, -0.5*d, 0.5*d, -0.27614*d, 0.5*d, 0,
                            "C", 0.5*d, 0.27614*d, 0.27614*d, 0.5*d, 0, 0.5*d,
                            "C", -0.27614*d, 0.5*d, -0.5*d, 0.27614*d, -0.5*d, 0];},

                'ellipse': function(width, height){
                            var w = width || 1,
                                h = w;
                            if ((typeof height === 'number')&&(height>0))
                            {
                              h = height;
                            }
                            return ["M", -0.5*w,0,
                            "C", -0.5*w, -0.27614*h, -0.27614*w, -0.5*h, 0, -0.5*h,
                            "C", 0.27614*w, -0.5*h, 0.5*w, -0.27614*h, 0.5*w, 0,
                            "C", 0.5*w, 0.27614*h, 0.27614*w, 0.5*h, 0, 0.5*h,
                            "C", -0.27614*w, 0.5*h, -0.5*w, 0.27614*h, -0.5*w, 0];},

                'square': function(width){
                            var w = width || 1;
                            return ['M', 0.5*w, -0.5*w, 'l', 0, w, -w, 0, 0, -w, 'z'];},

                'triangle': function(side){
                            var s = side || 1;
                            return ['M', 0.5*s, -0.289*s, 'l', -0.5*s, 0.866*s, -0.5*s, -0.866*s, 'z'];},

                'cross': function(width){
                            var w = width || 1;
                            return ['M', -0.5*w, 0, 'l', w, 0, 'M', 0, -0.5*w, 'l', 0, w];},

                'ex': function(diagonal){
                            var d = diagonal || 1;
                            return ['M', -0.3535*d,-0.3535*d, 'L',0.3535*d,0.3535*d,
                                    'M',-0.3535*d,0.3535*d, 'L',0.3535*d,-0.3535*d];}
                };
  }

  function DrawCmd(cmdStr, coords)   // canvas syntax draw commands
  {
    this.drawFn = cmdStr;       // String version of the canvas command to call
    this.parms = coords || [];  // world coordinates in [cp1x,cp1y, cp2x,cp2y, ... x,y]
    this.parmsPx = [];          // parms transformed into pixel coords
  }

  Cobj = function(commands, objtype, options)
  {
    var opt, prop, val;

    this.type = "PATH";             // enum of type to instruct the render method
    this.drawCmds = null;           // array of DrawCmd objects or an Img object or Text String
    this.bBoxCmds = [];             // DrawCmd array for the text or img bounding box
    this.dwgOrg = {x:0, y:0};       // drawing origin (0,0) may get translated
    this.width = 0;                 // IMG types use to store requested width (natural width used if none passed)
    this.height = 0;                //     "
    this.imgX = 0;                  // TEXT & IMG use these for obj.translate, obj.rotate, obj.scale
    this.imgY = 0;                  //     "
    this.imgLorgX = 0;              //     "
    this.imgLorgY = 0;              //     "
    this.imgXscale = 1;             //     "
    this.imgYscale = 1;             //     "
    this.imgDegs = 0;               //     "
    this.lorg = 1;                  // used by IMG type for temp storage of lorg while loading image
    this.dragNdrop = null;
    // properties set by setProperty if undefined render uses Cango default
    this.iso = false;               // true = maintain aspect ratio
    this.border = false;            // true = stroke outline with strokeColor & lineWidth
    this.strokeCol = null;          // renderer will stroke a path and shape outline in this color
    this.fillCol = null;            // only used if type = SHAPE
    this.lineWidth = 1;             // in case fat outline is wanted for Path or Shape outline
    this.lineCap = null;            // round butt or mitre
    this.fontSize = null;           // fontSize in points (TEXT only)
    this.fontWeight = null;         // fontWeight 100..900 (TEXT only)
    // drop shadow properties
    this.shadowOffsetX = 0;
    this.shadowOffsetY = 0;
    this.shadowBlur = 0;
    this.shadowColor = "#000000";

    if (typeof objtype === 'string')
    {
      if (['PATH', 'SHAPE', 'IMG', 'TEXT'].indexOf(objtype) !== -1)
      {
        this.type = objtype;
      }
    }
    if ((this.type === 'PATH')||(this.type === 'SHAPE'))
    {
      // send the Cgo2D (SVG) commands off to the canvas DrawCmd processor
      this.drawCmds = this.cgo2DtoDrawCmd(commands);
    }
    else if (this.type === 'IMG')
    {
      this.drawCmds = new Image();
      // start the download of the image
      this.drawCmds.src = commands;  // commands is the URL of the image to be loaded
    }
    else   // TEXT
    {
      this.drawCmds = commands;  // just store the Image or text String
    }

    opt = (typeof options === 'object')? options: {};   // avoid undeclared object errors
    // check for all supported options
    for (prop in opt)
    {
      // check that this is opt's own property, not inherited from prototype
      if (opt.hasOwnProperty(prop))
      {
        this.setProperty(prop, opt[prop]);
      }
    }
    // force iso true for TEXT and IMG
    if ((this.type === "IMG")||(this.type === "TEXT"))
    {
      this.iso = true;
    }
  }

  /* =============================================================================
   * Convert Cgo2D data array ['M',x,y, 'L',x1,y1, ... 'Q',cx,cy,x2,y2, 'A',r,r ...]
   * to array of canvas DrawCmd {drawFn:'moveTo', [x,y]}
   * ----------------------------------------------------------------------------*/
  Cobj.prototype.cgo2DtoDrawCmd = function(path, scl)
  {
    var x = 0,
        y = 0,
        px, py,
        c1x, c1y,
        rot, rx, ry, larc, swp,
        arc_segs,
        cmd, pc,
        cmdObj,
        seg, coords, lineSeg,
        commands = [],
        xScale = scl || 1,
        yScale = xScale,
        xOfs = 0,          // move the shape drawing origin
        yOfs = 0,
        i, j,
        segments = [];

    function segmentToBezier(cx, cy, th0, th1, rdx, rdy, sin_th, cos_th)
    {
      var a00 = cos_th * rdx,
          a01 = -sin_th * rdy,
          a10 = sin_th * rdx,
          a11 = cos_th * rdy,
          th_half = 0.5 * (th1 - th0),
          t = (8/3) * Math.sin(th_half * 0.5) * Math.sin(th_half * 0.5) / Math.sin(th_half),
          x1 = cx + Math.cos(th0) - t * Math.sin(th0),
          y1 = cy + Math.sin(th0) + t * Math.cos(th0),
          x3 = cx + Math.cos(th1),
          y3 = cy + Math.sin(th1),
          x2 = x3 + t * Math.sin(th1),
          y2 = y3 - t * Math.cos(th1);

      return [ a00 * x1 + a01 * y1, a10 * x1 + a11 * y1,
               a00 * x2 + a01 * y2, a10 * x2 + a11 * y2,
               a00 * x3 + a01 * y3, a10 * x3 + a11 * y3 ];
    }

    function arcToBezier(ox, oy, radx, rady, rotateX, large, sweep, x, y)
    {
      var th = rotateX * (Math.PI/180),
          sin_th = Math.sin(th),
          cos_th = Math.cos(th),
          rdx = Math.abs(radx),
          rdy = Math.abs(rady),
          qx, qy, ql,
          a00, a01, a10, a11,
          x0, y0, x1, y1,
          d,
          sfactor_sq,
          sfactor,
          xc, yc,
          th0, th1, th_arc,
          segmts,
          result = [],
          k, th2, th3;

      qx = cos_th * (ox - x) * 0.5 + sin_th * (oy - y) * 0.5;
      qy = cos_th * (oy - y) * 0.5 - sin_th * (ox - x) * 0.5;
      ql = (qx*qx) / (rdx*rdx) + (qy*qy) / (rdy*rdy);
      if (ql > 1)
      {
        ql = Math.sqrt(ql);
        rdx *= ql;
        rdy *= ql;
      }
      a00 = cos_th / rdx;
      a01 = sin_th / rdx;
      a10 = -sin_th / rdy;
      a11 = cos_th / rdy;
      x0 = a00 * ox + a01 * oy;
      y0 = a10 * ox + a11 * oy;
      x1 = a00 * x + a01 * y;
      y1 = a10 * x + a11 * y;
      d = (x1-x0) * (x1-x0) + (y1-y0) * (y1-y0);
      sfactor_sq = 1 / d - 0.25;
      if (sfactor_sq < 0)
      {
        sfactor_sq = 0;
      }
      sfactor = Math.sqrt(sfactor_sq);
      if (sweep == large)
      {
        sfactor = -sfactor;
      }
      xc = 0.5 * (x0 + x1) - sfactor * (y1-y0);
      yc = 0.5 * (y0 + y1) + sfactor * (x1-x0);
      th0 = Math.atan2(y0-yc, x0-xc);
      th1 = Math.atan2(y1-yc, x1-xc);
      th_arc = th1-th0;
      if (th_arc < 0 && sweep == 1)
      {
        th_arc += 2*Math.PI;
      }
      else if (th_arc > 0 && sweep == 0)
      {
        th_arc -= 2 * Math.PI;
      }
      segmts = Math.ceil(Math.abs(th_arc / (Math.PI * 0.5 + 0.001)));
      for (k=0; k<segmts; k++)
      {
        th2 = th0 + k * th_arc / segmts;
        th3 = th0 + (k+1) * th_arc / segmts;
        result.push(segmentToBezier(xc, yc, th2, th3, rdx, rdy, sin_th, cos_th));
      }

      return result;
    }

    if (!isArray(path))
    {
      return commands;
    }
    // break the array into command segments
    if (typeof path[0] === 'number')
    {
      // special case of data only array. Test if 1st element is a number then
      // treat array as 'M', x0, y0, 'L', x1, y1, x2, y2, ... ]
      segments[0] = ['M', path[0], path[1]];
      lineSeg = ['L'];
      for (j=2,i=2; i<path.length; i++)
      {
        if (typeof path[i] !== 'number')
        {
          break;
        }
      }
      segments[1] = lineSeg.concat(path.slice(j,i));
    }
    else
    {
      for(j=0, i=1; i<path.length; i++)
      {
        if (typeof path[i] == 'string')
        {
          segments.push(path.slice(j,i));
          j = i;
        }
      }
      segments.push(path.slice(j,i));    // push the last command out
    }

    for (i=0; i<segments.length; i++)
    {
      seg = segments[i];
      cmd = seg[0];
      if ((i==0)&&(cmd != 'M'))   // check that the first move is absolute
      {
        cmd = 'M';
      }
      coords = seg.slice(1);      // skip the command copy coords
      if (coords)
      {
        coords = coords.map(parseFloat);
      }
      switch(cmd)
      {
        case 'M':
          x = xOfs + xScale*coords[0];
          y = yOfs + yScale*coords[1];
          px = py = null;
          cmdObj = new DrawCmd('moveTo', [x, y]);
          commands.push(cmdObj);
          coords.splice(0, 2);      // delete the 2 coords from the front of the array
          while (coords.length>0)
          {
            x = xOfs + xScale*coords[0];            // eqiv to muliple 'L' calls
            y = yOfs + yScale*coords[1];
            cmdObj = new DrawCmd('lineTo', [x, y]); // any coord pair after first move is regarded as line
            commands.push(cmdObj);
            coords.splice(0, 2);
          }
          break;
        case 'm':
          x += xScale*coords[0];
          y += yScale*coords[1];
          px = py = null;
          cmdObj = new DrawCmd('moveTo', [x, y]);
          commands.push(cmdObj);
          coords.splice(0, 2);      // delete the 2 coords from the front of the array
          while (coords.length>0)
          {
            x += xScale*coords[0];                  // eqiv to muliple 'l' calls
            y += yScale*coords[1];
            cmdObj = new DrawCmd('lineTo', [x, y]); // any coord pair after first move is regarded as line
            commands.push(cmdObj);
            coords.splice(0, 2);
          }
          break;
        case 'L':
          while (coords.length>0)
          {
            x = xOfs + xScale*coords[0];
            y = yOfs + yScale*coords[1];
            cmdObj = new DrawCmd('lineTo', [x, y]);
            commands.push(cmdObj);
            coords.splice(0, 2);
          }
          px = py = null;
          break;
        case 'l':
          while (coords.length>0)
          {
            x += xScale*coords[0];
            y += yScale*coords[1];
            cmdObj = new DrawCmd('lineTo', [x, y]);
            commands.push(cmdObj);
            coords.splice(0, 2);
          }
          px = py = null;
          break;
        case 'H':
          x = xOfs + xScale*coords[0];
          px = py = null ;
          cmdObj = new DrawCmd('lineTo', [x, y]);
          commands.push(cmdObj);
          break;
        case 'h':
          x += xScale*coords[0];
          px = py = null ;
          cmdObj = new DrawCmd('lineTo', [x, y]);
          commands.push(cmdObj);
          break;
        case 'V':
          y = yOfs + yScale*coords[0];
          px = py = null;
          cmdObj = new DrawCmd('lineTo', [x, y]);
          commands.push(cmdObj);
          break;
        case 'v':
          y += yScale*coords[0];
          px = py = null;
          cmdObj = new DrawCmd('lineTo', [x, y]);
          commands.push(cmdObj);
          break;
        case 'C':
          while (coords.length>0)
          {
            c1x = xOfs + xScale*coords[0];
            c1y = yOfs + yScale*coords[1];
            px = xOfs + xScale*coords[2];
            py = yOfs + yScale*coords[3];
            x = xOfs + xScale*coords[4];
            y = yOfs + yScale*coords[5];
            cmdObj = new DrawCmd('bezierCurveTo', [c1x, c1y, px, py, x, y]);
            commands.push(cmdObj);
            coords.splice(0, 6);
          }
          break;
        case 'c':
          while (coords.length>0)
          {
            c1x = x + xScale*coords[0];
            c1y = y + yScale*coords[1];
            px = x + xScale*coords[2];
            py = y + yScale*coords[3];
            x += xScale*coords[4];
            y += yScale*coords[5];
            cmdObj = new DrawCmd('bezierCurveTo', [c1x, c1y, px, py, x, y]);
            commands.push(cmdObj);
            coords.splice(0, 6);
          }
          break;
        case 'S':
          if (px == null || !pc.match(/[sc]/i))
          {
            px = x;                // already absolute coords
            py = y;
          }
          cmdObj = new DrawCmd('bezierCurveTo', [x-(px-x), y-(py-y),
                                                xOfs + xScale*coords[0], yOfs + yScale*coords[1],
                                                xOfs + xScale*coords[2], yOfs + yScale*coords[3]]);
          commands.push(cmdObj);
          px = xOfs + xScale*coords[0];
          py = yOfs + yScale*coords[1];
          x = xOfs + xScale*coords[2];
          y = yOfs + yScale*coords[3];
          break;
        case 's':
          if (px == null || !pc.match(/[sc]/i))
          {
            px = x;
            py = y;
          }
          cmdObj = new DrawCmd('bezierCurveTo', [x-(px-x), y-(py-y),
                                                x + xOfs + xScale*coords[0], y + yOfs + yScale*coords[1],
                                                x + xOfs + xScale*coords[2], y + yOfs + yScale*coords[3]]);
          commands.push(cmdObj);
          px = x + xScale*coords[0];
          py = y + yScale*coords[1];
          x += xScale*coords[2];
          y += yScale*coords[3];
          break;
        case 'Q':
          px = xOfs + xScale*coords[0];
          py = yOfs + yScale*coords[1];
          x = xOfs + xScale*coords[2];
          y = yOfs + yScale*coords[3];
          cmdObj = new DrawCmd('quadraticCurveTo', [px, py, x, y]);
          commands.push(cmdObj);
          break;
        case 'q':
          cmdObj = new DrawCmd('quadraticCurveTo', [x + xScale*coords[0], y + yScale*coords[1],
                                                    x + xScale*coords[2], y + yScale*coords[3]]);
          commands.push(cmdObj);
          px = x + xScale*coords[0];
          py = y + yScale*coords[1];
          x += xScale*coords[2];
          y += yScale*coords[3];
          break;
        case 'T':
          if ((px == null) || (!pc.match(/[qt]/i)))
          {
            px = x;
            py = y;
          }
          else
          {
            px = x-(px-x);
            py = y-(py-y);
          }
          cmdObj = new DrawCmd('quadraticCurveTo', [px, py,
                                                    xOfs + xScale*coords[0], yOfs + yScale*coords[1]]);
          commands.push(cmdObj);
          px = x-(px-x);
          py = y-(py-y);
          x = xOfs + xScale*coords[0];
          y = yOfs + yScale*coords[1];
          break;
        case 't':
          if (px == null || !pc.match(/[qt]/i))
          {
            px = x;
            py = y;
          }
          else
          {
            px = x-(px-x);
            py = y-(py-y);
          }
          cmdObj = new DrawCmd('quadraticCurveTo', [px, py,
                                                    x + xScale*coords[0], y + yScale*coords[1]]);
          commands.push(cmdObj);
          x += xScale*coords[0];
          y += yScale*coords[1];
          break;
        case 'A':
          while (coords.length>0)
          {
            px = x;
            py = y;
            rx = xScale*coords[0];
            ry = xScale*coords[1];
            rot = -coords[2];          // rotationX: swap for CCW +ve
            larc = coords[3];          // large arc    should be ok
            swp = 1 - coords[4];       // sweep: swap for CCW +ve
            x = xOfs + xScale*coords[5];
            y = yOfs + yScale*coords[6];
            arc_segs = arcToBezier(px, py, rx, ry, rot, larc, swp, x, y);
            for (j=0; j<arc_segs.length; j++)
            {
              cmdObj = new DrawCmd('bezierCurveTo', arc_segs[j]);
              commands.push(cmdObj);
            }
            coords.splice(0, 7);
          }
          break;
        case 'a':
          while (coords.length>0)
          {
            px = x;
            py = y;
            rx = xScale*coords[0];
            ry = xScale*coords[1];
            rot = -coords[2];          // rotationX: swap for CCW +ve
            larc = coords[3];          // large arc    should be ok
            swp = 1 - coords[4];       // sweep: swap for CCW +ve
            x += xScale*coords[5];
            y += yScale*coords[6];
            arc_segs = arcToBezier(px, py, rx, ry, rot, larc, swp, x, y);
            for (j=0; j<arc_segs.length; j++)
            {
              cmdObj = new DrawCmd('bezierCurveTo', arc_segs[j]);
              commands.push(cmdObj);
            }
            coords.splice(0, 7);
          }
          break;
        case 'Z':
          cmdObj = new DrawCmd('closePath', []);
          commands.push(cmdObj);
          break;
        case 'z':
          cmdObj = new DrawCmd('closePath', []);
          commands.push(cmdObj);
          break;
      }
      pc = cmd;     // save the previous command for possible reflected control points
    }
    return commands;
  }

  Cobj.prototype.translate = function(x, y)
  {
    var cmd,
        i, j;

    if ((this.type === "IMG")||(this.type === "TEXT"))
    {
      // no points to shift just remember the offset to use when rendering
      this.imgX += x;   // IMG and TEXT types convert to pixels during render
      this.imgY -= y;   // IMG and TEXT Cobj.translate always in 'iso' will *xscl to get pixels, so flip sign here
      // now transform the text bounding box (just moveTo and lineTo, no cPts)
      for(j=0; j < this.bBoxCmds.length; j++)   // step through the draw segments
      {
        // check for ep since 'closePath' has no end point)
        if (this.bBoxCmds[j].parms.length)
        {
          this.bBoxCmds[j].parms[0] += x;
          this.bBoxCmds[j].parms[1] += y;
        }
      }
    }
    else
    {
      for (i=0; i<this.drawCmds.length; i++)
      {
        cmd = this.drawCmds[i];  // for clarity
        for (j=0; j<cmd.parms.length/2; j++)
        {
          cmd.parms[2*j] += x;
          cmd.parms[2*j+1] += y;
        }
      }
    }
  };

  Cobj.prototype.rotate = function(degs)
  {
    var cmd,
        A = Math.PI*degs/180.0,   // radians
        sinA = Math.sin(A),
        cosA = Math.cos(A),
        x, y, i, j;

    if ((this.type === "IMG")||(this.type === "TEXT"))
    {
      // no points to shift just remember the value to use when rendering
      this.imgDegs += degs;
    }
    else
    {
      for (i=0; i<this.drawCmds.length; i++)
      {
        cmd = this.drawCmds[i];  // for clarity
        for (j=0; j<cmd.parms.length/2; j++)
        {
          x = cmd.parms[2*j];
          y = cmd.parms[2*j+1];
          cmd.parms[2*j] = x*cosA - y*sinA;
          cmd.parms[2*j+1] = x*sinA + y*cosA;
        }
      }
    }
  };

  Cobj.prototype.scale = function(xScl, yScl, lineWdScl)
  {
    var cmd,
        xScale = xScl || 1,
        yScale = yScl || xScale,   // default to isotropic scaling
        lwScale = lineWdScl || 1,  // default to current value
        i, j;

    if ((this.type === "IMG")||(this.type === "TEXT"))
    {
      // no points to shift just remeber values to use when rendering
      this.imgXscale *= xScale;
      this.imgYscale *= yScale;
      this.imgX *= xScale;
      this.imgY *= yScale;
      // now transform the text bounding box (just moveTo and lineTo, no cPts)
      for(j=0; j < this.bBoxCmds.length; j++)   // step through the draw segments
      {
        // check for ep since 'closePath' has no end point)
        if (this.bBoxCmds[j].parms.length)
        {
          this.bBoxCmds[j].parms[0] *= xScale;
          this.bBoxCmds[j].parms[1] *= yScale;
        }
      }
    }
    else
    {
      for (i=0; i<this.drawCmds.length; i++)
      {
        cmd = this.drawCmds[i];  // for clarity
        for (j=0; j<cmd.parms.length/2; j++)
        {
          cmd.parms[2*j] *= xScale;
          cmd.parms[2*j+1] *= yScale;
        }
      }
      if (lwScale>0)    // only affects PATH and SHAPE types
      {
        this.lineWidth *= lwScale;
      }
    }
  };

  Cobj.prototype.appendPath = function(obj, delMove)
  {
    var dcs;

    if ((this.type === "SHAPE")||(this.type === "PATH"))
    {
      dcs = clone(obj.drawCmds);  // make new element not references
      if (delMove)  // delete the inital 'moveTo' command
      {
        this.drawCmds = this.drawCmds.concat(dcs.slice(1));
      }
      else
      {
        this.drawCmds = this.drawCmds.concat(dcs);
      }
    }
  };

  Cobj.prototype.revWinding = function()
  {
    // reverse the direction of drawing around a path, stops holes in shapes being filled
    var cmds,
        zCmd = null,
        revCmds = [],
        k, len,
        dParms, dCmd;

    function revPairs(ary)
    {
      if (ary.length === 2)
      {
        return ary;
      }
      var revAry = [];
      var i, j;
      for (i=0, j=ary.length/2-1; j>=0; i++, j--)
      {
        revAry[2*i] = ary[2*j];
        revAry[2*i+1] = ary[2*j+1];
      }
      return revAry;
    }

    if ((this.type === "IMG")||(this.type === "TEXT"))
    {
      return;
    }
    if (this.drawCmds[this.drawCmds.length-1].drawFn.toUpperCase() == "Z")
    {
      cmds = this.drawCmds.slice(0, -1);  // leave off Z
      zCmd = this.drawCmds.slice(-1);
    }
    else
    {
      cmds = this.drawCmds.slice(0);  // copy the whole array
    }
    // now step back along the path
    k = cmds.length-1;
    len = cmds[k].parms.length;
    dParms = [cmds[k].parms[len-2], cmds[k].parms[len-1]];
    dCmd = new DrawCmd("moveTo", dParms);
    revCmds.push(dCmd);
    cmds[k].parms = cmds[k].parms.slice(0,-2);  // weve used the last point so slice it off
    while (k>0)
    {
      dParms = revPairs(cmds[k].parms);
      len = cmds[k-1].parms.length;
      dParms.push(cmds[k-1].parms[len-2], cmds[k-1].parms[len-1]); // push last of next cmd
      dCmd = new DrawCmd(cmds[k].drawFn, dParms);
      revCmds.push(dCmd);
      cmds[k-1].parms = cmds[k-1].parms.slice(0,-2);  // weve used the last point so slice it off
      k--;
    }
    // add the 'z' if it was a closed path
    if (zCmd)
    {
      revCmds.push(zCmd);
    }

    this.drawCmds = revCmds;
  };

  Cobj.prototype.setDropShadow = function(xOffset, yOffset, blurRadius, shadowColor)
  {
    var xOfs = xOffset || 0,
        yOfs = yOffset || 0,
        radius = blurRadius || 0,
        color = shadowColor || "#000000";

    if (arguments.length)
    {
      this.shadowOffsetX = xOfs;
      this.shadowOffsetY = yOfs;
      this.shadowBlur = radius;
      this.shadowColor = color;
    }
    else     // setDropShadow() will clear dropShadow settings
    {
      this.shadowOffsetX = 0;
      this.shadowOffsetY = 0;
      this.shadowBlur = 0;
      this.shadowColor = "#000000";
    }
  };

  Cobj.prototype.setBorder = function(color, borderWidth)
  {
    if (arguments.length)
    {
      this.setProperty("border", true);
      this.setProperty("strokeColor", color);
      this.setProperty("lineWidth", borderWidth);
    }
    else     // setBorder() will result in no border
    {
      this.setProperty("border", false);
    }
  };

  Cobj.prototype.setProperty = function(propertyName, value)
  {
    if ((typeof propertyName !== "string")||(value === undefined)||(value == null))
    {
      return;
    }

    switch (propertyName.toLowerCase())
    {
      case "fillcolor":
        this.fillCol = value;
        break;
      case "strokecolor":
        this.strokeCol = value;
        break;
      case "linewidth":
      case "strokewidth":                 // for backward compatability
        this.lineWidth = Math.abs(value);
        break;
      case "linecap":
        if (typeof value != "string")
        {
          return;
        }
        if ((value == "butt")||(value =="round")||(value == "square"))
        {
          this.lineCap = value;
        }
        break;
      case "iso":
        if ((value === true)||(value === 'iso')||(value === 'isotropic'))
        {
          this.iso = true;
        }
        break;
      case "border":
        if (value === true)
        {
          this.border = true;
        }
        if (value === false)
        {
          this.border = false;
        }
        break;
      case "fontsize":
        this.fontSize = Math.abs(value);
        break;
      case "fontweight":
        if ((typeof value == "string")||((typeof value == "number")&&(value>=100)&&(value<=900)))
        {
          this.fontWeight = value;
        }
        break;
      case "imgwidth":
        this.width = Math.abs(value);
        break;
      case "imgheight":
        this.height = Math.abs(value);
        break;
      case "lorg":
        if ([1, 2, 3, 4, 5, 6, 7, 8, 9].indexOf(value) !== -1)
        {
          this.lorg = value;
        }
        break;
      default:
        return;
    }
  };

  Cobj.prototype.dup = function()
  {
    var newObj = new Cobj();

    newObj.type = this.type;
    newObj.drawCmds = clone(this.drawCmds);
    newObj.bBoxCmds = clone(this.bBoxCmds);
    newObj.dwgOrg = clone(this.dwgOrg);
    newObj.iso = this.iso;
    newObj.border = this.border;
    newObj.strokeCol = this.strokeCol;
    newObj.fillCol = this.fillCol;
    newObj.lineWidth = this.lineWidth;
    newObj.lineCap = this.lineCap;
    newObj.width = this.width;
    newObj.height = this.height;
    newObj.imgX = this.imgX;
    newObj.imgY = this.imgY;
    newObj.imgLorgX = this.imgLorgX;
    newObj.imgLorgY = this.imgLorgY;
    newObj.imgXscale = this.imgXscale;
    newObj.imgYscale = this.imgYscale;
    newObj.imgDegs = this.imgDegs;
    newObj.fontSize = this.fontSize;
    newObj.fontWeight = this.fontWeight;
    newObj.shadowOffsetX = this.shadowOffsetX;
    newObj.shadowOffsetY = this.shadowOffsetY;
    newObj.shadowBlur = this.shadowBlur;
    newObj.shadowColor = this.shadowColor;

    return newObj;
  };

//===============================================================================

  function Layer(canvasID, canvasElement)
  {
    this.id = canvasID;
    this.cElem = canvasElement;
    this.dragObjects = [];
  }

  function getLayer(cgo)
  {
    var i, lyr = cgo.bkgCanvas.layers[0];

    for (i=1; i < cgo.bkgCanvas.layers.length; i++)
    {
      if (cgo.bkgCanvas.layers[i].id === cgo.cId)
      {
        lyr = cgo.bkgCanvas.layers[i];
        break;
      }
    }
    return lyr;    // Layer object
  }

  Cango = function(canvasId)
  {
    var savThis = this,
        bkgId, bkgL;

    function resizeLayers()
    {
      var j, ovl,
          t = savThis.bkgCanvas.offsetTop + savThis.bkgCanvas.clientTop,
          l = savThis.bkgCanvas.offsetLeft + savThis.bkgCanvas.clientLeft,
          w = savThis.bkgCanvas.offsetWidth,
          h = savThis.bkgCanvas.offsetHeight;

      for (j=1; j<savThis.bkgCanvas.layers.length; j++)  // bkg is layer[0]
      {
        ovl = savThis.bkgCanvas.layers[j].cElem;
        if (ovl)
        {
          ovl.style.top = t+'px';
          ovl.style.left = l+'px';
          ovl.style.width = w+'px';
          ovl.style.height = h+'px';
        }
      }
    }

    this.cId = canvasId;
    this.cnvs = document.getElementById(canvasId);
    if (this.cnvs == null)
    {
      alert("can't find canvas "+canvasId);
      return;
    }
    this.bkgCanvas = this.cnvs;  // this is a background canvas so bkgCanvas points to itself
    // check if this is a context for an overlay
    if (canvasId.indexOf("_ovl_") !== -1)
    {
      // this is an overlay. get a reference to the backGround canvas
      bkgId = canvasId.slice(0,canvasId.indexOf("_ovl_"));
      this.bkgCanvas = document.getElementById(bkgId);
    }
    this.rawWidth = this.cnvs.offsetWidth;
    this.rawHeight = this.cnvs.offsetHeight;
    this.aRatio = this.rawWidth/this.rawHeight;
    if (!this.bkgCanvas.hasOwnProperty('layers'))
    {
      // create an array to hold all the overlay canvases for this canvas
      this.bkgCanvas.layers = [];
      // make a Layerobject for the bkgCanvas
      bkgL = new Layer(this.cId, this.cnvs);
      this.bkgCanvas.layers[0] = bkgL;
    }
    if ((typeof Timeline !== "undefined")&& !this.bkgCanvas.hasOwnProperty('timeline'))
    {
      // create an array to hold all the overlay canvases for this bkgCanavs
      this.bkgCanvas.timeline = new Timeline();
    }
    if (!this.cnvs.hasOwnProperty('resized'))
    {
      // make canvas native aspect ratio equal style box aspect ratio.
      // Note: rawWidth and rawHeight are floats, assignment to ints will truncate
      this.cnvs.setAttribute('width', this.rawWidth);    // reset canvas pixels width
      this.cnvs.setAttribute('height', this.rawHeight);  // don't use style for this
      this.cnvs.resized = true;
    }
    // this.buffered is set if the CangoAnimation module is loaded
    // this.buffered = true; // draw animation to off-screen buffer then bitblt complete frames
    // create off screen drawing buffer
    if (this.buffered)
    {
      this.cnvs.buf = document.createElement('canvas');      // create buffer in memory
      this.cnvs.buf.setAttribute('width', this.rawWidth);    // set number of graphics pixels
      this.cnvs.buf.setAttribute('height', this.rawHeight);  // to match screen canvas
      this.bufCtx = this.cnvs.buf.getContext('2d');  // animation drawing done off screen
    }
    this.ctx = this.cnvs.getContext('2d');    // draw direct to screen canvas
    this.vpW = this.rawWidth;         // vp width in pixels (default to full canvas size)
    this.vpH = this.rawHeight;        // vp height in pixels
    this.vpLLx = 0;                   // vp lower left from canvas left in pixels
    this.vpLLy = this.rawHeight;      // vp lower left from canvas top
    this.xscl = 1;                    // world x axis scale factor, default to pixels, +ve right
    this.yscl = -1;                   // world y axis scale factor, default to pixels, +ve up
    this.xoffset = 0;                 // world x origin offset from viewport left in pixels
    this.yoffset = 0;                 // world y origin offset from viewport bottom in pixels
    this.ctx.textAlign = "left";      // all offsets are handled in code so dragNdrop works
    this.ctx.textBaseline = "top";
    this.penCol = "rgba(0, 0, 0, 1.0)";         // black
    this.penWid = 1;                  // pixels
    this.lineCap = "butt";
    this.paintCol = "rgba(128, 128, 128, 1.0)"; // gray
    this.fontSize = 10;               // pt
    this.fontWeight = 400;            // 100..900, 400 = normal,700 = bold
    this.fontFamily = "Consolas, Monaco, 'Andale Mono', monospace";

    this.getUnique = function()
    {
      uniqueVal += 1;     // a private 'global'
      return uniqueVal;
    };

    this.initModules();    // this method may be re-defined by modules

    addEvent(window, 'resize', resizeLayers);
  };

  Cango.prototype.initModules = function(){};

  Cango.prototype.toViewportCoords = function(x, y)
  {
    // transform x,y in world coords to viewport coords (x axis 0 to 100, y axis same scale factor)
    var xPx = this.vpLLx+this.xoffset+x*this.xscl;
    var yPx = this.vpLLy+this.yoffset+y*this.yscl;

    return {x: 100*xPx/this.rawWidth, y: 100*(this.rawHeight-yPx)/this.rawWidth};
  };

  Cango.prototype.toPixelCoords = function(x, y)
  {
    // transform x,y in world coords to canvas pixel coords (top left is 0,0 y axis +ve down)
    var xPx = this.vpLLx+this.xoffset+x*this.xscl;
    var yPx = this.vpLLy+this.yoffset+y*this.yscl;

    return {x: xPx, y: yPx};
  };

  Cango.prototype.toWorldCoords = function(xPx, yPx)
  {
    // transform xPx,yPx in raw canvas pixels to world coords (lower left is 0,0 +ve up)
    var xW = (xPx - this.vpLLx - this.xoffset)/this.xscl;
    var yW = (yPx - this.vpLLy - this.yoffset)/this.yscl;

    return {x: xW, y: yW};
  };

  Cango.prototype.getCursorPosWC = function(evt)
  {
    // pass in any mouse event, returns the position of the cursor in raw pixel coords
    var e = evt||window.event;
    var rect = this.cnvs.getBoundingClientRect();

    var xW = (e.clientX - rect.left - this.vpLLx - this.xoffset)/this.xscl;
    var yW = (e.clientY - rect.top - this.vpLLy - this.yoffset)/this.yscl;

    return {x: xW, y: yW};
  };

  Cango.prototype.clearCanvas = function(fillColor)
  {
    var layerObj;

    if (fillColor)
    {
      this.ctx.fillStyle = fillColor;
      this.ctx.fillRect(0, 0, this.rawWidth, this.rawHeight);
    }
    else
    {
      this.ctx.clearRect(0, 0, this.rawWidth, this.rawHeight);
    }
    // all drawing erased, but graphics contexts remain intact
    // clear the dragObjects array, draggables put back when rendered
    layerObj = getLayer(this);
    layerObj.dragObjects.length = 0;
    // in case the CangoHTMLtext extension is used
    if (this.cnvs.alphaOvl && this.cnvs.alphaOvl.parentNode)
    {
      this.cnvs.alphaOvl.parentNode.removeChild(this.cnvs.alphaOvl);
    }
  };

  Cango.prototype.setViewport = function(lowerLeftX, lowerLeftY, w, h)
  {
    if (h && w && (h > 0) && (w > 0))
    {
      this.vpW = w*this.rawWidth/100;
      this.vpH = h*this.rawWidth/100;
      this.vpLLx = lowerLeftX*this.rawWidth/100;
      this.vpLLy = this.rawHeight-lowerLeftY*this.rawWidth/100;
    }
    else
    {
      this.vpW = this.rawWidth;
      this.vpH = this.rawHeight;
      this.vpLLx = 0;
      this.vpLLy = this.rawHeight;
    }
    this.setWorldCoords();     // if new viewport, world coords are garbage, so reset to defaults
  };

  Cango.prototype.fillViewport = function(fillColor)
  {
    var newCol = fillColor || this.paintCol;

    this.ctx.save();
    this.ctx.fillStyle = newCol;
    this.ctx.fillRect(this.vpLLx, this.vpLLy-this.vpH, this.vpW, this.vpH); // fill: top, left, wid, hgt
    this.ctx.restore();
  };

  Cango.prototype.setWorldCoords = function(lowerLeftX, lowerLeftY, spanX, spanY)
  {
    var vpLLxWC = lowerLeftX || 0,     // viewport lower left x in world coords
        vpLLyWC = lowerLeftY || 0;     // viewport lower left y in world coords

    if (spanX && (spanX > 0))
    {
      this.xscl = this.vpW/spanX;
    }
    else
    {
      this.xscl = 1;                   // use pixel units
    }
    if (spanY && (spanY > 0))
    {
      this.yscl = -this.vpH/spanY;
    }
    else
    {
      this.yscl = -this.xscl;          // square pixels
    }
    this.xoffset = -vpLLxWC*this.xscl;
    this.yoffset = -vpLLyWC*this.yscl;
  };

  Cango.prototype.setPropertyDefault = function(propertyName, value)
  {
    if ((typeof propertyName != "string")||(value == undefined)||(value == null))
    {
      return;
    }

    switch (propertyName.toLowerCase())
    {
      case "fillcolor":
        if ((typeof value == "string")||(typeof value == "object"))  // gradient will be an object
        {
          this.paintCol = value;
        }
        break;
      case "strokecolor":
        if ((typeof value == "string")||(typeof value == "object"))  // gradient will be an object
        {
          this.penCol = value;
        }
        break;
      case "linewidth":
      case "strokewidth":
        this.penWid = value;
        break;
      case "linecap":
        if ((typeof value == "string")&&((value == "butt")||(value =="round")||(value == "square")))
        {
          this.lineCap = value;
        }
        break;
      case "fontfamily":
        if (typeof value == "string")
        {
          this.fontFamily = value;
        }
        break;
      case "fontsize":
        this.fontSize = value;
        break;
      case "fontweight":
        if ((typeof value == "string")||((value >= 100)&&(value <= 900)))
        {
          this.fontWeight = value;
        }
        break;
      case "steptime":
        if ((value >= 15)&&(value <= 500))
        {
          this.stepTime = value;
        }
        break;
      default:
        return;
    }
  };

  Cango.prototype.linearGradientFill = function(x1, y1, x2, y2, x, y, scl, isotropic)
  {
    var xOfs = x || 0,
        yOfs = y || 0,
        xScl = scl || 1,
        yScl = scl || 1,
        p1x, p1y,
        p2x, p2y;

    if ((isotropic != undefined)&&(isotropic == 'iso'))
    {
      yScl *= -this.xscl/this.yscl;
    }
    // pixel version of world coordinate parms
    p1x = xOfs+x1*xScl;
    p1y = yOfs+y1*yScl;
    p2x = xOfs+x2*xScl;
    p2y = yOfs+y2*yScl;

    return this.ctx.createLinearGradient(this.xscl*p1x, this.yscl*p1y, this.xscl*p2x, this.yscl*p2y);
  };

  Cango.prototype.radialGradientFill = function(x1, y1, r1, x2, y2, r2, x, y, scl, isotropic)
  {
    var xOfs = x || 0,
        yOfs = y || 0,
        xScl = scl || 1,
        yScl = scl || 1,
        p1x, p1y, p1r,
        p2x, p2y, p2r;

    if ((isotropic != undefined)&&(isotropic == 'iso'))
    {
      yScl *= -this.xscl/this.yscl;
    }
    // world coordinate parms (equivalent to compile methods)
    p1x = xOfs+x1*xScl;
    p1y = yOfs+y1*yScl;
    p1r = r1*this.xscl*xScl;
    p2x = xOfs+x2*xScl;
    p2y = yOfs+y2*yScl;
    p2r = r2*this.xscl*xScl;

    return this.ctx.createRadialGradient(this.xscl*p1x, this.yscl*p1y, p1r, this.xscl*p2x, this.yscl*p2y, p2r);
  };

  Cango.prototype.dropShadow = function(obj, scl)
  {
    var xOfs = obj.shadowOffsetX || 0,
        yOfs = obj.shadowOffsetY || 0,
        radius = obj.shadowBlur || 0,
        color = obj.shadowColor || "#000000",
        xScale = scl || 1,
        yScale = scl || 1;

    if (this.ctx.shadowOffsetX !== undefined)     // check if supported
    {
      if ((obj.type === "SHAPE")||(obj.type === "PATH"))   // must scale for world coords (matrix scling not used)
      {
        xScale *= this.xscl;
        if (obj.iso)
        {
          yScale *= -this.xscl;
        }
        else
        {
          yScale *= this.yscl;
        }
      }
      else                         // no need to scale here (matrix scaling used)
      {
        xScale *= this.xscl;
        yScale *= -this.xscl;
      }

      this.ctx.shadowOffsetX = xOfs*xScale;
      this.ctx.shadowOffsetY = yOfs*yScale;
      this.ctx.shadowBlur = radius*xScale;
      this.ctx.shadowColor = color;
    }
  };

  Cango.prototype.render = function(pathObj, x, y, scl, degs)
  {
    var savThis = this,
        drawObjAry = [],
        i;

    // return a function to call so that 'i' value is frozen correctly
    function genCallback(obj, x, y, scl, degs)
    {
      return function(){configImgObj(obj, x, y, scl, degs);}
    }

    function configImgObj(imgObj, x, y, scl, degs)
    {
      var wid = imgObj.width || imgObj.drawCmds.width,    // default to natural width if none passed
          hgt = imgObj.height || wid*imgObj.drawCmds.height/imgObj.drawCmds.width,  // default keep aspect ratio
          wid2 = wid/2,
          hgt2 = hgt/2,
          dx = 0,
          dy = 0,
          ulx, uly, llx, lly, lrx, lry, urx, ury,
          lorgWC = [0, [0, 0],    [wid2, 0],   [wid, 0],
                       [0, hgt2], [wid2, hgt2], [wid, hgt2],
                       [0, hgt],  [wid2, hgt],  [wid, hgt]];

      if (lorgWC[imgObj.lorg] !== undefined)
      {
        dx = -lorgWC[imgObj.lorg][0];
        dy = -lorgWC[imgObj.lorg][1];
      }
      imgObj.imgLorgX = dx;     // world coords offset to drawing origin
      imgObj.imgLorgY = dy;
      imgObj.width = wid;   // in case it was not set and natural used
      imgObj.height = hgt;
      // construct the DrawCmds for the text bounding box
      ulx = imgObj.imgX+dx; uly = imgObj.imgY+dy;
      llx = imgObj.imgX+dx; lly = imgObj.imgY+dy+hgt;
      lrx = imgObj.imgX+dx+wid; lry = imgObj.imgY+dy+hgt;
      urx = imgObj.imgX+dx+wid; ury = imgObj.imgY+dy;
      imgObj.bBoxCmds[0] = new DrawCmd("moveTo", [ulx, -uly]);
      imgObj.bBoxCmds[1] = new DrawCmd("lineTo", [llx, -lly]);
      imgObj.bBoxCmds[2] = new DrawCmd("lineTo", [lrx, -lry]);
      imgObj.bBoxCmds[3] = new DrawCmd("lineTo", [urx, -ury]);
      imgObj.bBoxCmds[4] = new DrawCmd("closePath", []);

      savThis._paintImg(imgObj, x, y, scl, degs);
    }

    function configTextObj(txtObj)
    {
      var size = txtObj.fontSize || savThis.fontSize,     // fontSize in pts
          weight = txtObj.fontWeight || savThis.fontWeight,
          lorg = txtObj.lorg || 1,
          wid, hgt,   // Note: char cell is ~1.5* fontSize pixels high
          wid2, hgt2,
          lorgWC,
          dx = 0,
          dy = 0,
          ulx, uly, llx, lly, lrx, lry, urx, ury;

      // set the drawing context to measure the size
      savThis.ctx.save();
      savThis.ctx.font = size+"pt "+savThis.fontFamily;
      wid = savThis.ctx.measureText(txtObj.drawCmds).width;   // wid is in pixels
      savThis.ctx.restore();
      hgt = 1.5*size/1.3;   // Note: char cell height is ~1.5*fontSize in px, px size is 1.3*pt size
      wid /= savThis.xscl;  // convert to world coords
      hgt /= savThis.xscl;  // TEXT dimendions are 'iso' so use xscl
      wid2 = wid/2;
      hgt2 = hgt/2;
      lorgWC = [0, [0, 0],    [wid2, 0],   [wid, 0],
                   [0, hgt2], [wid2, hgt2], [wid, hgt2],
                   [0, hgt],  [wid2, hgt],  [wid, hgt]];
      if (lorgWC[lorg] !== undefined)
      {
        dx = -lorgWC[lorg][0];  // offset to drawing origin (converted to world coords)
        dy = -lorgWC[lorg][1];
      }
      txtObj.imgLorgX = dx;      // world coord offset to drawing origin
      txtObj.imgLorgY = dy;
      txtObj.width = wid;        // in world coords
      txtObj.height = hgt;
      // construct the DrawCmds for the text bounding box
      ulx = txtObj.imgX+dx; uly = txtObj.imgY+dy;
      llx = txtObj.imgX+dx; lly = txtObj.imgY+dy+hgt;
      lrx = txtObj.imgX+dx+wid; lry = txtObj.imgY+dy+hgt;
      urx = txtObj.imgX+dx+wid; ury = txtObj.imgY+dy;
      txtObj.bBoxCmds[0] = new DrawCmd("moveTo", [ulx, -uly]);
      txtObj.bBoxCmds[1] = new DrawCmd("lineTo", [llx, -lly]);
      txtObj.bBoxCmds[2] = new DrawCmd("lineTo", [lrx, -lry]);
      txtObj.bBoxCmds[3] = new DrawCmd("lineTo", [urx, -ury]);
      txtObj.bBoxCmds[4] = new DrawCmd("closePath", []);

      savThis._paintText(txtObj, x, y, scl, degs);
    }

    // task:function, ary: array with elements
  	function iterate(obj)
  	{
  	  var k, childNode;
      if (isArray(obj))          // array of Cobj or mixed Cobj and more arrays
      {
    		for (k=0; k<obj.length; k++)
    		{
    			childNode = obj[k];
      	  if (childNode.drawCmds !== undefined)    // test for Objs to draw
          {
            drawObjAry.push(childNode);       // just push into the array to be drawn
          }
          else if (isArray(childNode))
          {
    				iterate(childNode);
          }
    		}
      }
      else
      {
        drawObjAry.push(obj);
      }
  	}

    // handle Cobj or nested arrays of Objs
    iterate(pathObj);

    for (i=0; i<drawObjAry.length; i++)
    {
      if (drawObjAry[i].type === "IMG")
      {
        if (drawObjAry[i].drawCmds.width>0)  // image loaded and width set?
        {
          configImgObj(drawObjAry[i], x, y, scl, degs);
        }
        else
        {
          // Caution: genCallback is a closure to freeze the value of 'i'. It returns the callback fn
          addLoadEvent(drawObjAry[i].drawCmds, genCallback(drawObjAry[i], x, y, scl, degs));
          // check if it finished loading while we were setting up
          if (drawObjAry[i].drawCmds.width>0)  // image loaded and width set?
          {
            configImgObj(drawObjAry[i], x, y, scl, degs);
          }
        }
      }
      else if (drawObjAry[i].type == "TEXT")
      {
        configTextObj(drawObjAry[i]);
      }
      else    // PATH, SHAPE
      {
        this._paintPath(drawObjAry[i], x, y, scl, degs);
      }
    }
  };

  Cango.prototype._paintImg = function(pathObj, x, y, scl, degrees)
  {
    // should only be called after image has been loaded into drawCmds
    var A, sinA, cosA,
        img = pathObj.drawCmds,            // this is the place the image is stored in object
        xPos = x || 0,
        yPos = y || 0,
        xScale = scl || 1,
        scale = xScale*pathObj.imgXscale,  // imgXscale is from Cobj.scale (permanent)
        degs = degrees || 0,
        i, j, tp,
        currLr, aidx;

    function rotXY(x,y)
    {
      return [x*cosA - y*sinA, x*sinA + y*cosA];
    }

    this.ctx.save();   // save the clean ctx
    // set up dropShadow if any
    this.dropShadow(pathObj, xScale);
    // move the whole coordinate system to the xPos,yPos
    this.ctx.translate(this.vpLLx+this.xoffset+xPos*this.xscl, this.vpLLy+this.yoffset+yPos*this.yscl);
    degs += pathObj.imgDegs;
    if (degs)
    {
      A = degs*Math.PI/180.0;   // radians
      sinA = Math.sin(A);
      cosA = Math.cos(A);
      this.ctx.rotate(-A);   // rotate
    }
    // now insert the image origin with lorg offsets and scaled in width
    this.ctx.drawImage(img, this.xscl*scale*(pathObj.imgX+pathObj.imgLorgX),
                            this.xscl*scale*(pathObj.imgY+pathObj.imgLorgY),
                            this.xscl*scale*pathObj.width, this.xscl*scale*pathObj.height);

    // now transform the bounding box (skip final closePath, no cPts)
    for(j=0; j < pathObj.bBoxCmds.length-1; j++)   // step through the draw segments
    {
      if (degs)
      {
        tp = rotXY(pathObj.bBoxCmds[j].parms[0], pathObj.bBoxCmds[j].parms[1]);
      }
      else
      {
        tp = pathObj.bBoxCmds[j].parms;
      }
      pathObj.bBoxCmds[j].parmsPx[0] = this.vpLLx+this.xoffset+tp[0]*scale*this.xscl+xPos*this.xscl;
      pathObj.bBoxCmds[j].parmsPx[1] = this.vpLLy+this.yoffset-tp[1]*scale*this.xscl+yPos*this.yscl;
    }
    this.ctx.restore();    // undo the transforms (and drop shadow settings)
    if (pathObj.border)
    {
      this.ctx.beginPath();
      for (i=0; i < pathObj.bBoxCmds.length; i++)
      {
        this.ctx[pathObj.bBoxCmds[i].drawFn].apply(this.ctx, pathObj.bBoxCmds[i].parmsPx);
      }
      // if properties are undefined use Cango default
      this.ctx.strokeStyle = pathObj.strokeCol || this.penCol;
      this.ctx.lineWidth = (pathObj.lineWidth)? pathObj.lineWidth: this.penWid;
      this.ctx.lineCap = (pathObj.lineCap)? pathObj.lineCap: this.lineCap;
      this.ctx.stroke();
    }
    // save world coords of the drawing origin for drag n drop
    pathObj.dwgOrg.x = xPos;
    pathObj.dwgOrg.y = yPos;

    if (pathObj.dragNdrop != null)
    {
      // update dragNdrop layer to match this canavs
      currLr = getLayer(this);
      if (currLr !== pathObj.dragNdrop.layer)
      {
        if (pathObj.dragNdrop.layer)  // if not the first time rendered
        {
          // remove the object reference from the old layer
          aidx = pathObj.dragNdrop.layer.dragObjects.indexOf(this);
          if (aidx != -1)
          {
            pathObj.dragNdrop.layer.dragObjects.splice(aidx, 1);
          }
        }
      }
      pathObj.dragNdrop.cgo = this;
      pathObj.dragNdrop.layer = currLr;
      // now push it into Cango.dragObjects array, its checked by canvas mousedown event handler
      if (!pathObj.dragNdrop.layer.dragObjects.contains(pathObj))
      {
        pathObj.dragNdrop.layer.dragObjects.push(pathObj);
      }
    }
  };

  Cango.prototype._paintPath = function(pathObj, x, y, scl, degrees)
  {
    // used for type: PATH, SHAPE
    var A, sinA, cosA,
        xPos = x || 0,
        yPos = y || 0,
        scale = scl || 1,
        degs = degrees || 0,
        i, j, x1, y1,
        xTmp, yTmp,
        drawCmdsLen,
        parmsLen,
        xPx = this.vpLLx+this.xoffset+xPos*this.xscl,
        yPx = this.vpLLy+this.yoffset+yPos*this.yscl,
        xsl = this.xscl,
        ysl = pathObj.iso? -this.xscl : this.yscl,
        currLr, aidx;
    // don't use canvas matrix rotation or the gradient patterns will be rotated
    if (degs)
    {
      A = degs*Math.PI/180.0;   // radians
      sinA = Math.sin(A);
      cosA = Math.cos(A);
    }
    this.ctx.save();   // save current context
    // set up dropShadow if any
    this.dropShadow(pathObj, scale);
    // use canvas translation and scaling so gradient fill follow the object drawing origin and size
    this.ctx.translate(xPx, yPx);
    this.ctx.scale(scale, scale);
    // build the path by converting the world coord parms of each DrawCmd to parmPx
    // these can then be uses in drag and drop pointInpath testing
    this.ctx.beginPath();
    drawCmdsLen = pathObj.drawCmds.length;
    for (i=0; i < drawCmdsLen; i++)
    {
      parmsLen = pathObj.drawCmds[i].parms.length;
      for (j=0; j<parmsLen; j+=2)  // convert x,y coord pairs to pixel coords
      {
        xTmp = pathObj.drawCmds[i].parms[j];
        yTmp = pathObj.drawCmds[i].parms[j+1];
        if (degs)
        {
          x1 = xTmp*cosA - yTmp*sinA;
          y1 = xTmp*sinA + yTmp*cosA;
          xTmp = x1;
          yTmp = y1;
        }
        pathObj.drawCmds[i].parmsPx[j] = xTmp*xsl;
        pathObj.drawCmds[i].parmsPx[j+1] = yTmp*ysl;
      }
      this.ctx[pathObj.drawCmds[i].drawFn].apply(this.ctx, pathObj.drawCmds[i].parmsPx); // draw the path
    }
    if (pathObj.type === "SHAPE")      // if a SHAPE, fill with color
    {
      this.ctx.fillStyle = pathObj.fillCol || this.paintCol;
      this.ctx.fill();
    }
    if ((pathObj.type === "PATH")|| pathObj.border)
    {
      // clear drop shadow its done (and we might want to stroke border)
      this.ctx.shadowOffsetX = 0;
      this.ctx.shadowOffsetY = 0;
      this.ctx.shadowBlur = 0;
      // if properties are undefined use Cango default
      this.ctx.strokeStyle = pathObj.strokeCol || this.penCol;
      this.ctx.lineWidth = pathObj.lineWidth || this.penWid;
      this.ctx.lineCap = pathObj.lineCap || this.lineCap;
      this.ctx.stroke();
    }
    // undo the translation and scaling
    this.ctx.restore();

    // correct the pixel outline for any shift of drawing origin
    for (i=0; i < drawCmdsLen; i++)
    {
      parmsLen = pathObj.drawCmds[i].parms.length;
      for (j=0; j < parmsLen; j+=2)      // convert x,y coord pairs to world coords
      {
        pathObj.drawCmds[i].parmsPx[j] = pathObj.drawCmds[i].parmsPx[j]*scale + xPx;
        pathObj.drawCmds[i].parmsPx[j+1] = pathObj.drawCmds[i].parmsPx[j+1]*scale + yPx;
      }
    }
    // save world coords of the drawing origin for drag n drop
    pathObj.dwgOrg.x = xPos;
    pathObj.dwgOrg.y = yPos;
    if (pathObj.dragNdrop != null)
    {
      // update dragNdrop layer to match this canavs
      currLr = getLayer(this);
      if (currLr !== pathObj.dragNdrop.layer)
      {
        if (pathObj.dragNdrop.layer)  // if not the first time rendered
        {
          // remove the object reference from the old layer
          aidx = pathObj.dragNdrop.layer.dragObjects.indexOf(this);
          if (aidx != -1)
          {
            pathObj.dragNdrop.layer.dragObjects.splice(aidx, 1);
          }
        }
      }
      pathObj.dragNdrop.cgo = this;
      pathObj.dragNdrop.layer = currLr;
      // now push it into Cango.dragObjects array, its checked by canvas mousedown event handler
      if (!pathObj.dragNdrop.layer.dragObjects.contains(pathObj))
      {
        pathObj.dragNdrop.layer.dragObjects.push(pathObj);
      }
    }
  };

  Cango.prototype._paintText = function(pathObj, x, y, scl, degrees)
  {
    var A, sinA, cosA,
        xPos = x || 0,
        yPos = y || 0,
        xScale = scl || 1,
        scale = xScale*pathObj.imgXscale,  // imgXscale is from Cobj.scale (permanent)
        degs = degrees || 0,
        j, tp,
        fntWt, fntSz,
        currLr, aidx;

    function rotXY(x,y)
    {
      return [x*cosA - y*sinA, x*sinA + y*cosA];
    }

    // translate, rotate, scale must be handled at render time
    this.ctx.save();   // save current context
    // set up dropShadow if any
    this.dropShadow(pathObj, scale);
    // only the x, y passed to render are in non-iso coords (ie use yscl)
    this.ctx.translate(this.vpLLx+this.xoffset+xPos*this.xscl, this.vpLLy+this.yoffset+yPos*this.yscl);
    this.ctx.scale(scale, scale);
    this.ctx.save();   // save current rotation if any
    degs += pathObj.imgDegs;
    if (degs)
    {
      A = degs*Math.PI/180.0;   // radians
      sinA = Math.sin(A);
      cosA = Math.cos(A);
      this.ctx.rotate(-A);   // rotate
    }
    // if Cobj fontWeight or fontSize undefined use Cango default
    fntWt = (pathObj.fontWeight)? pathObj.fontWeight.toString(): this.fontWeight.toString();
    fntSz = (pathObj.fontSize)? pathObj.fontSize: this.fontSize;
    this.ctx.font = fntWt+" "+fntSz+"pt "+this.fontFamily;
    // set the fillStyle to obj.fillCol for text
    this.ctx.fillStyle = pathObj.fillCol || this.paintCol;
    // now actually fill the text (imgY is always 'iso' so *xscl)
    this.ctx.fillText(pathObj.drawCmds, this.xscl*(pathObj.imgX+pathObj.imgLorgX), this.xscl*(pathObj.imgY+pathObj.imgLorgY));
    if (pathObj.border)
    {
      // fill done, if dropShadow dont apply to the border (it will be on top of fill)
      this.ctx.shadowOffsetX = 0;
      this.ctx.shadowOffsetY = 0;
      this.ctx.shadowBlur = 0;
      // if properties are undefined use Cango default
      this.ctx.strokeStyle = pathObj.strokeCol || this.penCol;
      this.ctx.lineWidth = pathObj.lineWidth || this.penWid;
      this.ctx.lineCap = (pathObj.lineCap)? pathObj.lineCap: this.lineCap;
      // now actually fill the text
      this.ctx.strokeText(pathObj.drawCmds, this.xscl*(pathObj.imgX+pathObj.imgLorgX), this.xscl*(pathObj.imgY+pathObj.imgLorgY));
    }
    this.ctx.restore();   // undo the rotation
    // undo the translation and scaling
    this.ctx.restore();

    // now transform the bounding box (skip final closePath, no cPts)
    for(j=0; j < pathObj.bBoxCmds.length-1; j++)   // step through the draw segments
    {
      if (degs)
      {
        tp = rotXY(pathObj.bBoxCmds[j].parms[0], pathObj.bBoxCmds[j].parms[1]);
      }
      else
      {
        tp = pathObj.bBoxCmds[j].parms;
      }
      // parms for text are stored in pixels so don't scale bounding box by xscl
      pathObj.bBoxCmds[j].parmsPx[0] = this.vpLLx+this.xoffset+tp[0]*scale*this.xscl+xPos*this.xscl;
      pathObj.bBoxCmds[j].parmsPx[1] = this.vpLLy+this.yoffset-tp[1]*scale*this.xscl+yPos*this.yscl;
    }
    // save world coords of the drawing origin (often useful in drag n drop)
    pathObj.dwgOrg.x = xPos;
    pathObj.dwgOrg.y = yPos;
    if (pathObj.dragNdrop != null)
    {
      // update dragNdrop layer to match this canavs
      currLr = getLayer(this);
      if (currLr !== pathObj.dragNdrop.layer)
      {
        if (pathObj.dragNdrop.layer)  // if not the first time rendered
        {
          // remove the object reference from the old layer
          aidx = pathObj.dragNdrop.layer.dragObjects.indexOf(this);
          if (aidx != -1)
          {
            pathObj.dragNdrop.layer.dragObjects.splice(aidx, 1);
          }
        }
      }
      pathObj.dragNdrop.cgo = this;
      pathObj.dragNdrop.layer = currLr;
      // now push it into Cango.dragObjects array, its checked by canvas mousedown event handler
      if (!pathObj.dragNdrop.layer.dragObjects.contains(pathObj))
      {
        pathObj.dragNdrop.layer.dragObjects.push(pathObj);
      }
    }
  };

  Cango.prototype.compilePath = function(path, color, lineWidth, isotropic)
  {
    var lineWd = lineWidth || this.penWid,
        pathObj = new Cobj(path, "PATH", {"strokeColor":color, "lineWidth":lineWd, "iso":isotropic});

    return pathObj;
  };

  Cango.prototype.compileShape = function(path, color, isotropic)
  {
    var pathObj = new Cobj(path, "SHAPE", {"fillColor":color, "iso":isotropic});

    return pathObj;
  };

  Cango.prototype.compileText = function(str, color, fontSz, lorigin)
  {
    var txtObj = new Cobj(str, "TEXT", {"fillColor":color, "fontSize":fontSz, "lorg":lorigin});

    return txtObj;
  };

  Cango.prototype.compileImg = function(imgURL, w, lorigin)
  {
    var imgObj = new Cobj(imgURL, "IMG", {"imgWidth":w, "lorg":lorigin});  // imgObj.drawCmds = new image object

    return imgObj;   // image not loaded yet, this is the Cobj
  };

  Cango.prototype.drawPath = function(path, x, y, color, scl, isotropic)
  {
    var pathObj = this.compilePath(path, color, null, isotropic);
    this.render(pathObj, x, y, scl);

    return pathObj;
  };

  Cango.prototype.drawShape = function(path, x, y, fillColor, scl, isotropic)
  {
    // outline the same as fill color
    var pathObj = this.compileShape(path, fillColor, isotropic);
    this.render(pathObj, x, y, scl);

    return pathObj;
  };

  Cango.prototype.drawText = function(str, x, y, color, ptSize, lorg)
  {
    var pathObj = this.compileText(str, color, ptSize, lorg);
    this.render(pathObj, x, y);

    return pathObj;
  };

  Cango.prototype.drawImg = function(imgURL, x, y, w, lorigin)  // just load img then calls _paintImg
  {
    var imgObj = this.compileImg(imgURL, w, lorigin);
    this.render(imgObj, x, y);

    return imgObj;
  };

  Cango.prototype.clipPath = function(pathObj)
  {
    var pxlCoords = [],
        i, j;

    if ((pathObj.type == "IMG")||(pathObj.type == "TEXT"))
    {
      return;
    }
    this.ctx.save();             // not required when resetClip is supported
    this.ctx.beginPath();
    for (i=0; i<pathObj.drawCmds.length; i++)
    {
      for (j=0; j<pathObj.drawCmds[i].parms.length; j+=2)  // convert x,y coord pairs to pixel coords
      {
        pxlCoords[j] = this.vpLLx+this.xoffset+this.xscl*pathObj.drawCmds[i].parms[j];
        if (pathObj.iso)
        {
          pxlCoords[j+1] = this.vpLLy+this.yoffset-this.xscl*pathObj.drawCmds[i].parms[j+1];
        }
        else
        {
          pxlCoords[j+1] = this.vpLLy+this.yoffset+this.yscl*pathObj.drawCmds[i].parms[j+1];
        }
      }
      this.ctx[pathObj.drawCmds[i].drawFn].apply(this.ctx, pxlCoords); // actually draw the path
    }
    this.ctx.clip();
  };

  Cango.prototype.resetClip = function()
  {
    // this.ctx.resetClip();       // not supported in bowsers yet (Feb13)
    this.ctx.restore();            // use this until resetClip is supported
  };

  Cango.prototype.createLayer = function()
  {
    var ovlHTML, newCvs,
        w = this.rawWidth,
        h = this.rawHeight,
        unique, ovlId,
        nLyrs = this.bkgCanvas.layers.length,  // bkg is layer 0 so at least 1
        newL,
        topCvs;

    // do not create layers on overlays - only an background canvases
    if (this.cId.indexOf("_ovl_") !== -1)
    {
      // this is an overlay canvas - can't have overlays itself
      console.log("canvas layers can't create layers");
      return "";
    }

    unique = this.getUnique();
    ovlId = this.cId+"_ovl_"+unique;
    ovlHTML = "<canvas id='"+ovlId+"' style='position:absolute' width='"+w+"' height='"+h+"'></canvas>";
    topCvs = this.bkgCanvas.layers[nLyrs-1].cElem;  // eqv to this.cnvs.layers since only bkgCanavs can get here
    topCvs.insertAdjacentHTML('afterend', ovlHTML);
    newCvs = document.getElementById(ovlId);
    newCvs.style.backgroundColor = "transparent";
    newCvs.style.left = (this.bkgCanvas.offsetLeft+this.bkgCanvas.clientLeft)+'px';
    newCvs.style.top = (this.bkgCanvas.offsetTop+this.bkgCanvas.clientTop)+'px';
    // make it the same size as the background canvas
    newCvs.style.width = this.bkgCanvas.offsetWidth+'px';
    newCvs.style.height = this.bkgCanvas.offsetHeight+'px';
//    newCvs.style.pointerEvents = 'none';    // allow mouse events to pass down to bkgCanvas
    newL = new Layer(ovlId, newCvs);
    // save the ID in an array to facilitate removal
    this.bkgCanvas.layers.push(newL);

    return ovlId;    // return the new canvas id for call to new Cango(id)
  };

  Cango.prototype.deleteLayer = function(ovlyId)
  {
    var ovlNode, i;

    for (i=1; i<this.bkgCanvas.layers.length; i++)
    {
      if (this.bkgCanvas.layers[i].id === ovlyId)
      {
        ovlNode = this.bkgCanvas.layers[i].cElem;
        if (ovlNode)
        {
          // in case the CangoHTMLtext extension is used
          if (ovlNode.alphaOvl && ovlNode.alphaOvl.parentNode)
          {
            ovlNode.alphaOvl.parentNode.removeChild(ovlNode.alphaOvl);
          }
          ovlNode.parentNode.removeChild(ovlNode);
        }
        // now delete layers array element
        this.bkgCanvas.layers.splice(i,1);       // delete the id
      }
    }
  };

  Cango.prototype.deleteAllLayers = function()
  {
    var i, ovlNode;

    for (i = this.bkgCanvas.layers.length-1; i>0; i--)   // don't delete layers[0] its the bakg canavs
    {
      ovlNode = this.bkgCanvas.layers[i].cElem;
      if (ovlNode)
      {
        // in case the CangoHTMLtext extension is used
        if (ovlNode.alphaOvl && ovlNode.alphaOvl.parentNode)
        {
          ovlNode.alphaOvl.parentNode.removeChild(ovlNode.alphaOvl);
        }
        ovlNode.parentNode.removeChild(ovlNode);
      }
      // now delete layers array element
      this.bkgCanvas.layers.splice(i,1);
    }
  };

  // copy the basic graphics context values (for an overlay)
  Cango.prototype.dupCtx = function(src_graphCtx)
  {
    // copy all the graphics context parameters into the overlay ctx.
    this.vpLLx = src_graphCtx.vpLLx;      // vp lower left from canvas left in pixels
    this.vpLLy = src_graphCtx.vpLLy;      // vp lower left from canvas top
    this.xscl = src_graphCtx.xscl;        // world x axis scale factor
    this.yscl = src_graphCtx.yscl;        // world y axis scale factor
    this.xoffset = src_graphCtx.xoffset;  // world x origin offset from viewport left in pixels
    this.yoffset = src_graphCtx.yoffset;  // world y origin offset from viewport bottom in pixels
    this.penCol = src_graphCtx.penCol.slice(0);   // copy value not reference
    this.penWid = src_graphCtx.penWid;    // pixels
    this.lineCap = src_graphCtx.lineCap.slice(0);
    this.paintCol = src_graphCtx.paintCol.slice(0);
    this.fontFamily = src_graphCtx.fontFamily.slice(0);
    this.fontSize = src_graphCtx.fontSize;
    this.fontWeight = src_graphCtx.fontWeight;
  };

  Cango.prototype.toImgObj = function(obj)
  {
    // find the bounding box
    var top, rgt, bot, lft,
        dx, dy, w, h,
        xsc = this.xscl,
        ysc = obj.iso? -this.xscl : this.yscl,
        spr = {},
        img = new Image(),
        imgObj = new Cobj(img, "IMG"),
        i, j;

    function configImgObj()  // call when image loaded
    {
      var wid = imgObj.drawCmds.width,    // its loaded now so width is valid
          hgt = imgObj.drawCmds.height;  // keep aspect ratio

      imgObj.imgX = dx;     // world coords offset to drawing origin
      imgObj.imgY = dy-hgt;
      imgObj.width = wid;
      imgObj.height = hgt;
      // construct the DrawCmds for the text bounding box
      imgObj.bBoxCmds[0] = new DrawCmd("moveTo", [dx, -dy+hgt]);      // ul
      imgObj.bBoxCmds[1] = new DrawCmd("lineTo", [dx, -dy]);          // ll
      imgObj.bBoxCmds[2] = new DrawCmd("lineTo", [dx+wid, -dy]);      // lr
      imgObj.bBoxCmds[3] = new DrawCmd("lineTo", [dx+wid, -dy+hgt]);  // ur
      imgObj.bBoxCmds[4] = new DrawCmd("closePath", []);
    }

    if ((obj.type != 'PATH')&&(obj.type != 'SHAPE'))
    {
      return null;
    }
    // find pixel dimensions of the obj bounding box
    lft = rgt = obj.drawCmds[0].parms[0];
    bot = top = obj.drawCmds[0].parms[1];
    for (i=1; i < obj.drawCmds.length; i++)
    {
      for (j=0; j < obj.drawCmds[i].parms.length; j += 2)
      {
        // step through each draw command and find max and min end point or control point
        if (obj.drawCmds[i].parms[j] > rgt)
        {
          rgt = obj.drawCmds[i].parms[j];
        }
        if  (obj.drawCmds[i].parms[j] < lft)
        {
          lft = obj.drawCmds[i].parms[j];
        }
        if (obj.drawCmds[i].parms[j+1] > top)
        {
          top = obj.drawCmds[i].parms[j+1];
        }
        if  (obj.drawCmds[i].parms[j+1] < bot)
        {
          bot = obj.drawCmds[i].parms[j+1];
        }
      }
    }

    dx = lft*xsc-2;         // add a couple of pixels for aliasing
    dy = bot*ysc+2;
    w = (rgt - lft)*xsc+4;
    h = -(top - bot)*ysc+4;

    spr.width = w;        // pixels
    spr.height = h;
    spr.buf = document.createElement('canvas');    // create buffer in memory
    spr.buf.setAttribute('width', spr.width);
    spr.buf.setAttribute('height', spr.height);
    spr.gc = {};
    // now patch up a pseudo Cango context for _paintPath
    spr.gc.cId = 'sprite';
    spr.gc.cnvs = spr.buf;
    spr.gc.ctx = spr.gc.cnvs.getContext('2d');    // draw direct to screen canvas
    spr.gc.rawWidth = spr.width;          // width and height are like Image object
    spr.gc.rawHeight = spr.height;
    spr.gc.vpW = spr.gc.rawWidth;         // vp width in pixels (default to full canvas size)
    spr.gc.vpH = spr.gc.rawHeight;        // vp height in pixels
    spr.gc.vpLLx = 0;                     // vp lower left from canvas left in pixels
    spr.gc.vpLLy = spr.gc.rawHeight;      // vp lower left from canvas top
    spr.gc.xoffset = -dx;      // drawn at current pixel resolution
    spr.gc.yoffset = -dy;
    spr.gc.xscl = this.xscl;
    spr.gc.yscl = this.yscl;
    // use parent Cango values for drawing
    spr.gc.penCol = this.penCol.slice(0);   // copy value not reference
    spr.gc.penWid = this.penWid;    // pixels
    spr.gc.lineCap = this.lineCap.slice(0);
    spr.gc.paintCol = this.paintCol.slice(0);

    // render the obj
    this._paintPath.call(spr.gc, obj);  // paint obj dwgOrg at 0,0
    // load the buffer canvas as an image into an Cobj type 'IMG'
    imgObj.drawCmds.src = "";
    // img is rendered at pixel resolution (*xscl). The reference widht is in world coords
    // corrected by the scale factor 'imgXscale', so fix this for current world coords
    imgObj.imgXscale = 1/this.xscl;

    addLoadEvent(imgObj.drawCmds, configImgObj);
    // start to load the image
    imgObj.drawCmds.src = spr.gc.cnvs.toDataURL();

    return imgObj;
  };

  svgToCgo2D = function(svgPath, xRef, yRef)
  {
    var cgoData = [],
        segs, strs,
        cmd, seg, cmdLetters, crds,
        scaleX = 1,
        scaleY = -1,                      // flip all the y coords to +ve up
        ofsX = xRef || 0,                 // move the drawing reference point
        ofsY = yRef || 0,
        k;

    function segsToCgo2D(segs, xRef, yRef, xScl, yScl)
    {
      var x = 0,
          y = 0,
          px, py,
          c1x, c1y,
          rot, rx, ry, larc, swp,
          segmt, cmnd, pc,
          commands = [],
          xScale = xScl || 1,
          yScale = yScl || xScale,   // in case only single scale factor passed
          xOfs = xRef || 0,          // move the shape drawing origin
          yOfs = yRef || 0,
          i, coords;

      for (i=0; i<segs.length; i++)
      {
        segmt = segs[i];
        cmnd = segmt[0];
        if ((i==0)&&(cmnd != 'M'))   // check that the first move is absolute
        {
          cmnd = 'M';
        }
        coords = segmt.slice(1);      // skip the command copy coords
        if (coords)
        {
          coords = coords.map(parseFloat);
        }
        switch(cmnd)
        {
          case 'M':
            x = xOfs + xScale*coords[0];
            y = yOfs + yScale*coords[1];
            px = py = null;
            commands.push('M', x, y);
            coords.splice(0, 2);      // delete the 2 coords from the front of the array
            while (coords.length>0)
            {
              x = xOfs + xScale*coords[0];                // eqiv to muliple 'L' calls
              y = yOfs + yScale*coords[1];
              commands.push('L', x, y);
              coords.splice(0, 2);
            }
            break;
          case 'm':
            x += xScale*coords[0];
            y += yScale*coords[1];
            px = py = null;
            commands.push('M', x, y);
            coords.splice(0, 2);      // delete the 2 coords from the front of the array
            while (coords.length>0)
            {
              x += xScale*coords[0];              // eqiv to muliple 'l' calls
              y += yScale*coords[1];
              commands.push('L', x, y);
              coords.splice(0, 2);
            }
            break;
          case 'L':
            while (coords.length>0)
            {
              x = xOfs + xScale*coords[0];
              y = yOfs + yScale*coords[1];
              commands.push('L', x, y);
              coords.splice(0, 2);
            }
            px = py = null;
            break;
          case 'l':
            while (coords.length>0)
            {
              x += xScale*coords[0];
              y += yScale*coords[1];
              commands.push('L', x, y);
              coords.splice(0, 2);
            }
            px = py = null;
            break;
          case 'H':
            x = xOfs + xScale*coords[0];
            px = py = null ;
            commands.push('L', x, y);
            break;
          case 'h':
            x += xScale*coords[0];
            px = py = null ;
            commands.push('L', x, y);
            break;
          case 'V':
            y = yOfs + yScale*coords[0];
            px = py = null;
            commands.push('L', x, y);
            break;
          case 'v':
            y += yScale*coords[0];
            px = py = null;
            commands.push('L', x, y);
            break;
          case 'C':
            while (coords.length>0)
            {
              c1x = xOfs + xScale*coords[0];
              c1y = yOfs + yScale*coords[1];
              px = xOfs + xScale*coords[2];
              py = yOfs + yScale*coords[3];
              x = xOfs + xScale*coords[4];
              y = yOfs + yScale*coords[5];
              commands.push('C', c1x, c1y, px, py, x, y);
              coords.splice(0, 6);
            }
            break;
          case 'c':
            while (coords.length>0)
            {
              c1x = x + xScale*coords[0];
              c1y = y + yScale*coords[1];
              px = x + xScale*coords[2];
              py = y + yScale*coords[3];
              x += xScale*coords[4];
              y += yScale*coords[5];
              commands.push('C', c1x, c1y, px, py, x, y);
              coords.splice(0, 6);
            }
            break;
          case 'S':
            if (px == null || !pc.match(/[sc]/i))
            {
              px = x;                // already absolute coords
              py = y;
            }
            commands.push('C', x-(px-x), y-(py-y),
                              xOfs + xScale*coords[0], yOfs + yScale*coords[1],
                              xOfs + xScale*coords[2], yOfs + yScale*coords[3]);
            px = xOfs + xScale*coords[0];
            py = yOfs + yScale*coords[1];
            x = xOfs + xScale*coords[2];
            y = yOfs + yScale*coords[3];
            break;
          case 's':
            if (px == null || !pc.match(/[sc]/i))
            {
              px = x;
              py = y;
            }
            commands.push('C', x-(px-x), y-(py-y),
                              x + xOfs + xScale*coords[0], y +yOfs + yScale*coords[1],
                              x + xOfs + xScale*coords[2], y +yOfs + yScale*coords[3]);
            px = x + xScale*coords[0];
            py = y + yScale*coords[1];
            x += xScale*coords[2];
            y += yScale*coords[3];
            break;
          case 'Q':
            px = xOfs + xScale*coords[0];
            py = yOfs + yScale*coords[1];
            x = xOfs + xScale*coords[2];
            y = yOfs + yScale*coords[3];
            commands.push('Q', px, py, x, y);
            break;
          case 'q':
            commands.push('Q', [x + xScale*coords[0], y + yScale*coords[1],
                                x + xScale*coords[2], y + yScale*coords[3]]);
            px = x + xScale*coords[0];
            py = y + yScale*coords[1];
            x += xScale*coords[2];
            y += yScale*coords[3];
            break;
          case 'T':
            if (px == null || !pc.match(/[qt]/i))
            {
              px = x;
              py = y;
            }
            else
            {
              px = x-(px-x);
              py = y-(py-y);
            }
            commands.push('Q', px, py, xOfs + xScale*coords[0], yOfs + yScale*coords[1]);
            px = x-(px-x);
            py = y-(py-y);
            x = xOfs + xScale*coords[0];
            y = yOfs + yScale*coords[1];
            break;
          case 't':
            if (px == null || !pc.match(/[qt]/i))
            {
              px = x;
              py = y;
            }
            else
            {
              px = x-(px-x);
              py = y-(py-y);
            }
            commands.push('Q', px, py, x + xScale*coords[0], y + yScale*coords[1]);
            x += xScale*coords[0];
            y += yScale*coords[1];
            break;
          case 'A':
            while (coords.length>0)
            {
              px = x;
              py = y;
              rx = xScale*coords[0];
              ry = xScale*coords[1];
              rot = coords[2];            // don't switch to CCW +ve cgo2DToDrawCmd will
              larc = coords[3];
              swp = coords[4];            // sweep: don't switch swap to CCW +ve
              x = xOfs + xScale*coords[5];
              y = yOfs + yScale*coords[6];
              commands.push('A', rx, ry, rot, larc, swp, x, y);
              coords.splice(0, 7);
            }
            break;
          case 'a':
            while (coords.length>0)
            {
              px = x;
              py = y;
              rx = xScale*coords[0];
              ry = xScale*coords[1];
              rot = coords[2];          // don't switch to CCW +ve cgo2DToDrawCmd will
              larc = coords[3];
              swp = coords[4];          // sweep: don't switch swap to CCW +ve
              x += xScale*coords[5];
              y += yScale*coords[6];
              commands.push('A', rx, ry, rot, larc, swp, x, y);
              coords.splice(0, 7);
            }
            break;
          case 'Z':
            commands.push('Z');
            break;
          case 'z':
            commands.push('Z');
            break;
        }
        pc = cmnd;     // save the previous command for possible reflected control points
      }
      return commands;
    }

    if (typeof svgPath != 'string')
    {
      return cgoData;
    }
    // this is a preprocessor to get an svg Path string into 'Cgo2D' format
    segs = [];
    strs = svgPath.split(/(?=[a-df-z])/i);  // avoid e in exponents
    // now array of strings with command letter start to each
    for (k=0; k<strs.length; k++)
    {
      seg = strs[k];
      // get command letter into an array
      cmdLetters = seg.match(/[a-z]/i);
      if (!cmdLetters)
      {
        return cgoData;
      }
      cmd = cmdLetters.slice(0,1);
      if ((k==0)&&(cmd[0] != 'M'))   // check that the first move is absolute
      {
        cmd[0] = 'M';
      }
      crds = seg.match(/[\-\+]?[0-9]*\.?[0-9]+([eE][\-\+]?[0-9]+)?/gi);
      if (crds)
      {
        crds = crds.map(parseFloat);
      }
      segs.push(cmd.concat(crds));
    }

    // now send these off to the svg segs to Cgo2D
    cgoData = segsToCgo2D(segs, ofsX, -ofsY, scaleX, scaleY);

    return cgoData;  // array in Cgo2D format ['M', x, y, 'L', x, y, ....]
  };

  return Cango;
}());
